// -*- C++ -*-
/*!
 * @file StringUtil.h
 * @brief String operation utility
 * @date $Date: 2007-12-31 03:08:07 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2003-2008
 *     Noriaki Ando
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: StringUtil.h 826 2008-08-26 08:13:39Z n-ando $
 *
 */

#ifndef COIL_STRINGUTIL_H
#define COIL_STRINGUTIL_H

#include <string>
#include <vector>
#include <sstream>

namespace coil
{
  typedef std::vector<std::string> vstring;

  /*!
   * @if jp
   * @brief string  wstring ւ̕ϊ
   *
   * ^ꂽ string  wstring ɕϊ
   *
   * @param str std::string ^͕̓
   * @return std::wstring ^̏o͕
   *
   * @else
   * @brief string to wstring conversion
   *
   * This function convert from a string to a wstring.
   * 
   * @param str The input std::string type string
   * @return Converted std::wstring type string
   *
   * @endif
   */
  std::wstring string2wstring(std::string str);

  /*!
   * @if jp
   * @brief wstring  string ւ̕ϊ
   *
   * ^ꂽ wstring  string ɕϊ
   *
   * @param str std::wstring ^͕̓
   * @return std::string ^̏o͕
   *
   * @else
   * @brief wstring to string conversion
   *
   * This function convert from a wstring to a string.
   * 
   * @param str The input std::wstring type string
   * @return Converted std::string type string
   *
   * @endif
   */
  std::string wstring2string(std::wstring wstr);

  /*!
   * @if jp
   * @brief 啶ւ̕ϊ
   *
   * ^ꂽ啶ɕϊ
   *
   * @param str ͕
   *
   * @else
   * @brief Uppercase String Transformation
   *
   * This function transforms a given string to uppercase letters
   * 
   * @param str The input string
   *
   * @endif
   */
  void toUpper(std::string& str);

  /*!
   * @if jp
   * @brief ւ̕ϊ
   *
   * ^ꂽɕϊ
   *
   * @param str ͕
   *
   * @else
   * @brief Lowercase String Transformation
   *
   * This function transforms a given string to lowercase letters
   * 
   * @param str The input string
   *
   * @endif
   */
  void toLower(std::string& str);

  /*!
   * @if jp
   * @brief ̓Xg[1sǂݍ
   *
   * ̓Xg[1sǂݍށB
   * ̓Xg[͂̉sR[hUNIX, Windows̉sR[ĥꂩA
   * ͍݂ĂĂ悢B
   *
   * @param istr ̓Xg[
   * @param line ǂݍ񂾕i[ϐ
   *
   * @return sǂݍ񂾕̒
   *
   * @else
   * @brief Read a line from input stream
   *
   * This function reads a line from input stream.
   * UNIX, Windows or mixed line feed code is acceptable.
   *
   * @param istr The input stream.
   * @param line The output variable to store string to be read.
   *
   * @return The length of read string except line feed character.
   *
   * @endif
   */
  int getlinePortable(std::istream& istr, std::string& line);
  
  /*!
   * @if jp
   * @brief 񂪃GXP[vĂ邩f
   *
   * w肳ꂽGXP[vĂ邩ǂ𔻒fB
   *
   * @param str GXP[vĂ邩ǂf镶܂ޕ
   * @param pos GXP[vĂ邩ǂf镶̈ʒu
   *
   * @return w肵GXP[vĂ true, ȊO false
   *
   * @else
   * @brief Check whether the character is escaped or not
   *
   * Check whether the specified character is escaped or not
   *
   * @param str The string that includes the character to be investigated.
   * @param pos The position of the character to be investigated.
   *
   * @return True if the specified character is escaped, else False.
   *
   * @endif
   */
  bool isEscaped(const std::string& str, std::string::size_type pos);
  
  /*!
   * @if jp
   * @brief GXP[v
   *
   * ̕GXP[vV[PXɕϊB<br>
   * HT -> "\t" <br>
   * LF -> "\n" <br>
   * CR -> "\r" <br>
   * FF -> "\f" <br>
   * VONI[gA_uNI[gɂĂ͂Ƃɏ͂ȂB
   *
   * @param str GXP[vΏە
   *
   * @return GXP[vʕ
   *
   * @else
   *
   * @brief Escape string
   *
   * The following characters are converted. <br>
   * HT -> "\t" <br>
   * LF -> "\n" <br>
   * CR -> "\r" <br>
   * FF -> "\f" <br>
   * Single quote and double quote are not processed.
   *
   * @param str The target string for the escape
   *
   * @return Result string of the escape
   *
   * @endif
   */
  std::string escape(const std::string str);
  
  /*!
   * @if jp
   * @brief ̃GXP[v߂
   *
   * ̃GXP[vV[PX𕶎ɕϊB<br>
   * "\t" -> HT <br>
   * "\n" -> LF <br>
   * "\r" -> CR <br>
   * "\f" -> FF <br>
   * "\"" -> "  <br>
   * "\'" -> '  <br>
   * GXP[v̊Sȋtϊɂ͂ȂĂȂ߁AӂKvB
   *
   * @param str AGXP[vΏە
   *
   * @return AGXP[vʕ
   *
   * @else
   *
   * @brief Unescape string
   *
   * The following characters are converted. <br>
   * "\t" -> HT <br>
   * "\n" -> LF <br>
   * "\r" -> CR <br>
   * "\f" -> FF <br>
   * "\"" -> "  <br>
   * "\'" -> '  <br>
   * Note: This is not complete inversion of the escape processing.
   *
   * @param str The target string for the unescape
   *
   * @return Result string of the unescape
   *
   * @endif
   */
  std::string unescape(const std::string str);

  /*!
   * @if jp
   * @brief ̋󔒕폜
   *
   * ^ꂽ̋󔒕폜B
   * 󔒕ƂĈ̂' '(Xy[X)'\\t'(^u)B
   *
   * @param str 󔒕폜
   *
   * @else
   * @brief Erase blank characters of string
   *
   * Erase blank characters that exist at the head of the given string.
   * Space ' 'and tab '\\t' are supported as the blank character.
   *
   * @param str The target blank characters of string for the erase
   *
   * @endif
   */
  void eraseBlank(std::string& str);

  /*!
   * @if jp
   * @brief ̐擪̋󔒕폜
   *
   * ^ꂽ̐擪ɑ݂󔒕폜B
   * 󔒕ƂĈ̂' '(Xy[X)'\\t'(^u)B
   *
   * @param str 擪󔒕폜
   *
   * @else
   * @brief Erase the head blank characters of string
   *
   * Erase the blank characters that exist at the head of the given string.
   * Space ' 'and tab '\\t' are supported as the blank character.
   *
   * @param str The target head blank characters of string for the erase
   *
   * @endif
   */
  void eraseHeadBlank(std::string& str);
  
  /*!
   * @if jp
   * @brief ̖̋󔒕폜
   *
   * ^ꂽ̖ɑ݂󔒕폜B
   * 󔒕ƂĈ̂' '(Xy[X)'\\t'(^u)B
   *
   * @param str 󔒕폜
   *
   * @else
   * @brief Erase the tail blank characters of string
   *
   * Erase the blank characters that exist at the tail of the given
   * string.  Space ' 'and tab '\\t' are supported as the blank
   * character.
   *
   * @param str The target tail blank characters of string for the erase
   *
   * @endif
   */
  void eraseTailBlank(std::string& str);

  /*!
   * @if jp
   * @brief ̐擪E̋󔒕폜
   *
   * ^ꂽ̐擪іɑ݂󔒕폜B
   * 󔒕ƂĈ̂' '(Xy[X)'\\t'(^u)B
   *
   * @param str 擪󔒕폜
   *
   * @else
   * @brief Erase the head blank and the tail blank characters of string
   *
   * Erase the head blank characters and the blank characters that
   * exist at the tail of the given string.  Space ' 'and tab '\\t'
   * are supported as the blank character.
   *
   * @param str The target tail blank characters of string for the erase
   *
   * @endif
   */
  void eraseBothEndsBlank(std::string& str);

  /*!
   * @if jp
   * @brief ̂𐳋K
   *
   * ^ꂽ̐擪іɑ݂󔒕폜A
   * pׂďɕϊB
   *
   * @param str Ώە
   *
   * @else
   * @brief Erase the head/tail blank and replace upper case to lower case
   *
   * Erase the head blank characters and the blank characters that
   * exist at the tail of the given string.  Space ' 'and tab '\\t'
   * are supported as the blank character.
   * And all upper case cahracters are converted into lower case.
   *
   * @param str The target string for the erase
   *
   * @endif
   */
  std::string normalize(std::string& str);
  
  /*!
   * @if jp
   * @brief u
   *
   * ^ꂽɑ΂āAw肵̒usB
   *
   * @param str uΏە
   * @param from u
   * @param to u敶
   *
   * @else
   * @brief Replace string
   *
   * Replace the given string with the specified characters.
   *
   * @param str The target characters of string for replacement processing
   * @param from Characters of replacement source
   * @param to Characters of replacement destination
   *
   * @endif
   */
  void replaceString(std::string& str, const std::string from,
                     const std::string to);
  
  /*!
   * @if jp
   * @brief 𕪊ŕ
   * 
   * ݒ肳ꂽ^ꂽf~^ŕB
   *
   * @param input Ώە
   * @param delimiter (f~^)
   *
   * @return 񕪊ʃXg
   *
   * @else
   * @brief Split string by delimiter
   * 
   * Split the set string by the given delimiter
   *
   * @param input The target characters of string for split
   * @param delimiter Split string (delimiter)
   *
   * @return Split string result list
   *
   * @endif
   */
  vstring split(const std::string& input,
                const std::string& delimiter,
                bool ignore_empty = false);
  
  /*!
   * @if jp
   * @brief ^ꂽboollɕϊ
   * 
   * w肳ꂽAtrue\Afalse\ƔrǍʂ
   * boollƂĕԂB
   * řʁAtrue\Afalse\̂ǂƂvȂꍇ́A
   * ^ꂽftHglԂB
   *
   * @param str fΏە
   * @param yes true\
   * @param no false\
   * @param default_value ftHgl(ftHgl:true)
   * @else
   * @brief Convert given string into bool value
   * 
   * Compare the specified string with the true representation string and
   * the false representation string, and return the result as bool value.
   * If it matches neither the true representation string nor the false
   * representation string as a result of the comparison, the given default
   * value will be return.
   *
   * @param str The target string for investigation
   * @param yes The true representation string
   * @param no The false representation string
   * @param default_value The default value (The default value:true)
   * @endif
   */
  bool toBool(std::string str, std::string yes, std::string no, 
              bool default_value = true);
  /*!
   * @if jp
   * @brief 񃊃Xgɂ镶񂪊܂܂邩ǂ
   * 
   * 1ɃJ}؂̃XgA2ɒTΏەw肵A
   * ̕񂪑1̒Ɋ܂܂邩𔻒fB
   *
   * @param list ΏۃXg
   * @param value T
   * @return true: ܂܂Afalse: ܂܂Ȃ
   *
   * @else
   * @brief Include if a string is included in string list
   * 
   * if the second argument is included in the comma separated string
   * list of the first argument, This operation returns "true value".
   *
   * @param list The target comma separated string
   * @param value The searched string
   * @return true: included, false: not included
   *
   * @endif
   */
  bool includes(const vstring& list, std::string value,
                bool ignore_case = true);
  bool includes(const std::string& list, std::string value,
                bool ignore_case = true);
  
  /*!
   * @if jp
   * @brief ^ꂽ񂪐΃pXǂ𔻒f
   *
   * ^ꂽ񂪐΃pX\ł邩ǂ𔻒fB
   * 񂪈ȉ̏ꍇɂ͐΃pXƂĔfB
   *  - 擪'/' (UNIX̏ꍇ)
   *  - 擪RAt@xbg{'/'{'\\' (Windows̏ꍇ)
   *  - 擪Q'\\\\' (Windowslbg[NpX̏ꍇ)
   *
   * @param str Ώە
   *
   * @return ΃pX茋
   *
   * @else
   * @brief Investigate whether the given string is absolute path or not
   *
   * Investigate whether the given string is absolute path or not.
   * Investigate it as an absolute path, if the string is as follows:
   *  - The first character '/' (UNIX)
   *  - The first 3 characters are alphabet {'/'{'\\' (Windows)
   *  - The first 2 characters are '\\\\' (Windows network path)
   *
   * @param str The target string for the investigation
   *
   * @return Investigation result of absolute path
   *
   * @endif
   */
  bool isAbsolutePath(const std::string& str);
  
  /*!
   * @if jp
   * @brief ^ꂽURLǂ𔻒f
   *
   * ^ꂽURL\ǂ𔻒fB
   * ^ꂽ񒆂ɁA'://'Ƃ񂪊܂܂Ăꍇɂ
   * URL\ƂĔfB
   *
   * @param str Ώە
   *
   * @return URL茋
   *
   * @else
   * @brief Investigate whether the given string is URL or not
   *
   * Investigate whether the given string is URL or not.
   * When the string '://' is included in the given character string,
   * make it of URL representation.
   *
   * @param str The target string for investigation
   *
   * @return URL investigation result
   *
   * @endif
   */
  bool isURL(const std::string& str);
  
  /*!
   * @if jp
   * @brief ^ꂽIuWFNgstd::stringɕϊ
   *
   * Ŏw肳ꂽIuWFNg𕶎ɕϊB
   *
   * @param n ϊΏۃIuWFNg
   *
   * @return ϊ
   *
   * @else
   * @brief Convert the given object to std::string
   *
   * Convert the object specified by the argument to string.
   *
   * @param n The target object for conversion
   *
   * @return String conversion result
   *
   * @endif
   */
  template <class Printable>
  std::string otos(Printable n)
  {
    std::stringstream str_stream;
    str_stream << n;
    return str_stream.str();
  };
  
  /*!
   * @if jp
   * @brief ^ꂽstd::stringIuWFNgɕϊ
   *
   * ŗ^ꂽw肳ꂽIuWFNgɕϊB
   *
   * @param val ϊIuWFNg
   * @param str ϊ
   *
   * @return ϊs
   *
   * @else
   * @brief Convert the given std::string to object.
   *
   * Convert string given by the argument to specified object.
   *
   * @param val The object of conversion destination
   * @param str String of conversion source
   *
   * @return Conversion processing result
   *
   * @endif
   */
  template <typename To>
  bool stringTo(To& val, const char* str)
  {
    if (str == 0) { return false; }

    std::stringstream s;
    if ((s << str).fail()) { return false; }
    if ((s >> val).fail()) { return false; }
    return true;
  }
  
  template<>
  bool stringTo<std::string>(std::string& val, const char* str);

  /*!
   * @if jp
   * @brief ^ꂽ񃊃Xgd폜
   *
   * ŗ^ꂽ񃊃Xgd폜Xg쐬B
   *
   * @param sv mF񃊃Xg
   *
   * @return d폜ʃXg
   *
   * @else
   * @brief Eliminate duplication from the given string list
   *
   * Create a list of eliminating duplication from the string list
   * given by the argument.
   *
   * @param sv The string list for confirmation source
   *
   * @return Eliminating duplication result list
   *
   * @endif
   */
  vstring unique_sv(vstring sv);
  
  /*!
   * @if jp
   * @brief ^ꂽ񃊃XgCSV𐶐
   *
   * ŗ^ꂽ񃊃Xg̊evfׂCSV𐶐B
   * 񃊃Xg̏ꍇɂ͋󔒕ԂB
   *
   * @param sv CSVϊΏە񃊃Xg
   *
   * @return CSVϊʕ
   *
   * @else
   * @brief Create CSV file from the given string list
   *
   * Create CSV that arranged each element of the character string
   * list given by the argument.
   * If the string list is empty, the null will be returned.
   *
   * @param sv The target string list for creating CSV
   *
   * @return String of CSV creating result
   *
   * @endif
   */
  std::string flatten(vstring sv);
  
  /*!
   * @if jp
   * @brief ^ꂽ񃊃XgXgɕϊ
   *
   * ŗ^ꂽ񃊃Xg̊evf'\\0'AXg
   * ɕϊB
   *
   * @param args ϊΏە񃊃Xg
   *
   * @return ϊʕ
   *
   * @else
   * @brief Convert the given string list into the argument list
   *
   * Convert the string list into the argument list by adding '\\0' to
   * each element at the end of the string list given by the argument
   *
   * @param args The target string list for conversion
   *
   * @return String of argument conversion result
   *
   * @endif
   */
  char** toArgv(const vstring& args); 


  std::string sprintf(char const * __restrict fmt, ...);
  
}; // namepsace coil
#endif // COIL_STRINGUTIL_H
